/* $Id: kprintf.c,v 1.6 2002/08/23 20:23:49 ken Exp $
	silly test for using/understanding varargs.
	originally started from manpage for varargs
	then turned into a homegrown printf, then ported to avr
*/

#include <string.h>
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <pgmspace.h>
#include "kprintf.h"
#include "uart.h"


/* this is all spelled out,  because i can't use PSTR here, 
	because it puts {}'s around itself, which causes gcc to pitch a fit */
static const char hex[] __attribute__((progmem)) = "0123456789ABCDEF";


/* XXX optimization to do: merge the signed and unsigned,
	but need to find a way to do casting unsigned/signed, etc */

/****************
	PRINTSIGNED
	takes care of signed number printing for kprintf
	XXX the original uses consts here. why? 
********************/
static void
printSigned(int dec, unsigned char maxbuf, unsigned char base)
{
	char dbuf[MAXDIGITS]  = { '\0' };  
	int dlen = 0;
	if (dec < 0 ) {
		UartSend("-", 1, BSSMEM);
		dec = -dec; /* i need positive numbers from here on */
	}
	do {
		dbuf[sizeof(dbuf) - dlen  - 1] = PRG_RDB(&hex[ dec % base ]);
		dec = dec/base;
		dlen++;
	} while(dec && dlen < maxbuf);
	UartSend(&dbuf[sizeof(dbuf) - dlen ], dlen, BSSMEM);

} /* END PRINTSIGNED */

/****************
	PRINTUNSIGNED
	does unsigned hex. really can be inlined.
********************/
static void
printUnsigned(unsigned int dec, unsigned char maxbuf)
{
	char dbuf[MAXDIGITS]  = { '\0' };  
	int dlen = 0;
	do {
		dbuf[sizeof(dbuf) - dlen  - 1] = PRG_RDB(&hex[ dec % 16 ]);
		dec = dec/16;
		dlen++;
	} while(dec && dlen < maxbuf);
	UartSend(&dbuf[sizeof(dbuf) - dlen ], dlen, BSSMEM);

} /* END PRINTUNSIGNED */



/****************
	KPRINTF
	a cheap, small, simple printf for embedded use
********************/
int
_kprintf(char * fmt, ...)
{
	va_list ap;
	char * fptr = NULL;
	char ch;

	va_start(ap, fmt);
	fptr = fmt;
	for (ch = PRG_RDB(fptr); ch != '\0'; ch = PRG_RDB(++fptr)){
		if(ch != '%'){
			UartSend(fptr, 1, PM);
		} else {
			ch = PRG_RDB(++fptr);

			/* deal with max digit specs */
			int maxChar = 0;
			while(ch >= '0' && ch <= '9'){
				/* maxChar *= 10; if MAXDIGITS > 10 */
				maxChar += ch - '0';
				ch = PRG_RDB(++fptr);
			}
			/* string from sram */
			if(ch == 's') { 
				char  *s;
				int slen;
				s = va_arg(ap, char *);
				/* from uclibc, cheaper than strlen */
				/* XXX some kind of protection against runaway strings? */
				for(slen = 0; s[slen]; slen++) { /* so gcc doesn't nuke */ };
				UartSend(s, slen, BSSMEM);
			
			/* string from pgstr */
			} else if(ch == 'S') { 
				char  *s;
				int slen = 0;
				s = va_arg(ap, char *);
				while( PRG_RDB(&s[slen]) != '\0') {
					slen++;
				}
				UartSend(s, slen, PM);

			/* char */
			} else if(ch == 'c') { 
				/* va_arg doesn't do char's? */
				char cc = (char) va_arg(ap, int);
				UartSend( &cc, 1, BSSMEM);

			/* signed int */
			} else if(ch == 'd') { 
				/* from etherboot */
				int dec = va_arg(ap, int);
				printSigned(dec, maxChar ? maxChar : MAXDIGITS, DEC);

			/* signed hex   XXX am i sure about this? signed hex?? */
			} else if(ch == 'x'){
				int dec = va_arg(ap, int);
				printSigned(dec, maxChar ? maxChar : XDIGITS, HEX );

			/* unsigned hex  */
			} else if(ch == 'X'){
				unsigned int dec = va_arg(ap, unsigned int);
				printUnsigned(dec, maxChar ? maxChar : XDIGITS);
			} /* end if-chain */
		} /* end % */
	}
	va_end(ap);

	return(0);
} /* END KPRINTF */

/* EOF */

